import datetime
import os

import pytest

from zmail.exceptions import InvalidArguments, ZmailInternalError
from zmail.helpers import (convert_date_to_datetime, encode_mail_header,
                           first_not_none, get_abs_path, get_intersection,
                           make_address_header, make_list, match_conditions)
from zmail.structures import CaseInsensitiveDict


def test_convert_date_to_datetime():
    now = datetime.datetime.now()
    now_year = now.year
    now_month = now.month
    now_day = now.day

    assert now is convert_date_to_datetime(now)

    with pytest.raises(InvalidArguments):
        convert_date_to_datetime('test-with-invalidArgs')

    datetime_as_str = ''
    res_datetime = convert_date_to_datetime(datetime_as_str)
    assert res_datetime.year == now_year
    assert res_datetime.month == now_month
    assert res_datetime.day == now_day

    for datetime_as_str in ('1000', '1000-'):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == now_month
        assert res_datetime.day == now_day
        assert res_datetime.hour == 0
        assert res_datetime.minute == 0
        assert res_datetime.second == 0

    for datetime_as_str in ('1000-05', '1000-5', '1000-05-', '1000-5-'):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == 5
        assert res_datetime.day == now_day
        assert res_datetime.hour == 0
        assert res_datetime.minute == 0
        assert res_datetime.second == 0

    for datetime_as_str in ('1000-10-01', '1000-10-1'):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == 10
        assert res_datetime.day == 1
        assert res_datetime.hour == 0
        assert res_datetime.minute == 0
        assert res_datetime.second == 0

    for datetime_as_str in ('1000-10-01 03', '1000-10-01 3:'):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == 10
        assert res_datetime.day == 1
        assert res_datetime.hour == 3
        assert res_datetime.minute == 0
        assert res_datetime.second == 0

    for datetime_as_str in ('1000-10-01 04:05', '1000-10-01 4:5:'):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == 10
        assert res_datetime.day == 1
        assert res_datetime.hour == 4
        assert res_datetime.minute == 5
        assert res_datetime.second == 0

    for datetime_as_str in ('1000-10-01 02:03:04', '1000-10-01 2:3:4',
                            '1000-10-01 02:03:04', '1000-10-01 2:3:4 '):
        res_datetime = convert_date_to_datetime(datetime_as_str)
        assert res_datetime.year == 1000
        assert res_datetime.month == 10
        assert res_datetime.day == 1
        assert res_datetime.hour == 2
        assert res_datetime.minute == 3
        assert res_datetime.second == 4


def test_match_conditions():
    mock_mail_headers = CaseInsensitiveDict({
        'subject': 'zmail',
        'from': 'zmail_test@gmail.com',
        'date': datetime.datetime(year=2000, month=1, day=15)
    })

    headers_without_date = mock_mail_headers.copy()
    headers_without_date['date'] = None

    headers_without_subject = mock_mail_headers.copy()
    headers_without_subject['subject'] = None

    headers_without_from = mock_mail_headers.copy()
    headers_without_from['from'] = None

    assert match_conditions(mock_mail_headers)
    assert match_conditions(mock_mail_headers, subject='zmail')
    assert match_conditions(mock_mail_headers, sender='zmail')
    assert match_conditions(mock_mail_headers, subject='zmail', sender='zmail')

    assert match_conditions(mock_mail_headers, subject='123') is False
    assert match_conditions(mock_mail_headers, subject='123', sender='zmail') is False

    assert match_conditions(mock_mail_headers, end_time=datetime.datetime.now())
    assert match_conditions(mock_mail_headers, end_time=datetime.datetime(2018, 1, 1))
    assert match_conditions(mock_mail_headers, start_time=datetime.datetime.now()) is False
    assert match_conditions(mock_mail_headers, start_time=datetime.datetime(2018, 1, 1)) is False

    assert match_conditions(mock_mail_headers, start_time=datetime.datetime(2000, 1, 1),
                            end_time=datetime.datetime.now())
    assert match_conditions(mock_mail_headers, start_time=datetime.datetime(2000, 1, 1),
                            end_time=datetime.datetime(2018, 1, 1))

    assert match_conditions(headers_without_date, start_time=datetime.datetime.now()) is False
    assert match_conditions(headers_without_date, end_time=datetime.datetime.now()) is False
    assert match_conditions(headers_without_date, start_time=datetime.datetime(2000, 1, 1),
                            end_time=datetime.datetime.now()) is False

    assert match_conditions(headers_without_subject, subject='test') is False

    assert match_conditions(headers_without_from, sender='test') is False


def test_get_intersection():
    assert get_intersection((1, 5), (1, 5)) == [1, 2, 3, 4, 5]
    assert get_intersection((1, 5), (None, None)) == [1, 2, 3, 4, 5]

    assert get_intersection((1, 5), (2, 3)) == [2, 3]
    assert get_intersection((1, 5), (3, None)) == [3, 4, 5]
    assert get_intersection((1, 5), (None, 3)) == [1, 2, 3]

    assert get_intersection((1, 5), (-3, None)) == [1, 2, 3, 4, 5]
    assert get_intersection((1, 5), (None, 10)) == [1, 2, 3, 4, 5]

    assert get_intersection((0, 0), (1, 2)) == []
    assert get_intersection((0, 0), (None, None)) == [0]
    assert get_intersection((1, 0), (1, 2)) == []


def test_encode_mail_header():
    assert encode_mail_header('') == ''
    assert encode_mail_header('你好') == '=?utf-8?b?5L2g5aW9?='
    assert encode_mail_header('123') == '=?utf-8?b?MTIz?='


def test_make_list():
    assert make_list('') == ['']
    assert make_list(1) == [1]
    assert make_list((1,)) == [(1,)]

    assert make_list(list()) == list()

    assert make_list([1, 2, 3]) == [1, 2, 3]


def test_make_address_header():
    assert make_address_header([]) == ''

    t = ['123@123.com']
    assert make_address_header(t) == '<123@123.com>'

    t = [('你好', '123@123.com')]
    assert make_address_header(t) == '=?utf-8?b?5L2g5aW9?= <123@123.com>'

    t = ['123@123.com', ('你好', '123@123.com')]
    assert make_address_header(t) == '<123@123.com>, =?utf-8?b?5L2g5aW9?= <123@123.com>'

    t = t[::-1]
    assert make_address_header(t) == '=?utf-8?b?5L2g5aW9?= <123@123.com>, <123@123.com>'


def test_first_not_none():
    with pytest.raises(ZmailInternalError):
        first_not_none()
        first_not_none(None)
    assert first_not_none(True) is True
    assert first_not_none(None, False, True) is False
    assert first_not_none(True, None) is True


def test_get_abs_path():
    here = os.path.abspath(os.path.dirname(__file__))
    assert os.path.exists(get_abs_path(os.path.join(here, 'test_helpers.py')))



import logging
import smtplib
from unittest import mock

import pytest

from zmail.info import get_supported_server_info
from zmail.server import SMTPServer


@pytest.fixture
def smtp_server(accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    config = get_supported_server_info(username)
    return SMTPServer(username, password,
                      host=config['smtp_host'],
                      port=config['smtp_port'],
                      ssl=config['smtp_ssl'],
                      tls=config['smtp_tls'],
                      timeout=60,
                      debug=False)


@pytest.fixture
def smtp_server_config(accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    config = get_supported_server_info(username)
    return {
        'username': username,
        'password': password,
        'host': config['smtp_host'],
        'port': config['smtp_port'],
        'ssl': config['smtp_ssl'],
        'tls': config['smtp_tls'],
        'timeout': 60,
        'debug': False
    }


def test_smtp_server_init(smtp_server_config, accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    new_logger = mock.Mock()
    srv = SMTPServer(**smtp_server_config, log=new_logger)
    assert srv.username == username
    assert srv.password == password
    assert srv.host == smtp_server_config['host']
    assert srv.port == smtp_server_config['port']
    assert srv.ssl == smtp_server_config['ssl']
    assert srv.tls == smtp_server_config['tls']
    assert srv.timeout == 60
    assert srv.debug is False
    assert srv.log is new_logger


def test_smtp_server_default_logger(smtp_server_config):
    srv = SMTPServer(**smtp_server_config)
    assert srv.log == logging.getLogger('zmail')


def test_smtp_ssl_tls_error(smtp_server_config):
    configs = smtp_server_config.copy()
    configs.update(tls=True)
    with pytest.raises(TypeError):
        SMTPServer(**configs)


def test_smtp_make_server(smtp_server_config):
    srv = SMTPServer(**smtp_server_config)
    srv._make_server()
    assert isinstance(srv.server, (smtplib.SMTP_SSL if srv.ssl else smtplib.SMTP))


def test_smtp_login(smtp_server_config):
    srv = SMTPServer(**smtp_server_config)
    srv.login()
    assert srv.is_login() is True and srv._login is True
    srv.logout()


def test_smtp_duplicate_login(smtp_server_config):
    srv = SMTPServer(**smtp_server_config)
    srv.login()
    srv.login()
    assert srv.is_login() is True and srv._login is True
    srv.logout()


def test_smtp_logout(smtp_server_config):
    srv = SMTPServer(**smtp_server_config)
    srv.login()
    srv.logout()
    assert srv.is_login() is False and srv._login is False


def test_smtp_context(smtp_server):
    assert smtp_server.is_login() is False
    with smtp_server as server:
        assert isinstance(server, SMTPServer)
        assert server.server is not None
        assert server.is_login() is True and server._login is True
    assert server.server is None
    assert server.is_login() is False and server._login is False


def test_smtp_check_available(smtp_server_config):
    assert SMTPServer(**smtp_server_config).check_available()
    incorrect_config = smtp_server_config.copy()
    incorrect_config.update(host='')
    assert SMTPServer(**incorrect_config).check_available() is False


# Methods

def test_send_mail():
    pass



"""
zmail.server
~~~~~~~~~~~~
This module provides a MailServer object to communicate with mail server.
"""

import datetime
import logging
import poplib
import smtplib
import warnings
from typing import Iterable, List, Optional

from .abc import BaseServer
from .exceptions import InvalidArguments
from .helpers import (convert_date_to_datetime, first_not_none,
                      get_intersection, make_address_header, make_list,
                      match_conditions)
from .mime import Mail
from .parser import parse_headers, parse_mail
from .settings import __local__
from .structures import CaseInsensitiveDict

# Fix poplib bug.
poplib._MAXLINE = 4096

logger = logging.getLogger('zmail')


class MailServer:

    def __init__(self, username: str, password: str,
                 smtp_host: str, smtp_port: int,
                 pop_host: str, pop_port: int,
                 smtp_ssl: bool, pop_ssl: bool,
                 smtp_tls: bool, pop_tls: bool,
                 debug: bool = False, log=None, timeout=60,
                 auto_add_from=True, auto_add_to=True):
        self.username = username
        self.password = password
        self.debug = debug
        self.log = log or logger
        self.timeout = timeout

        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_ssl = smtp_ssl
        self.smtp_tls = smtp_tls

        self.pop_host = pop_host
        self.pop_port = pop_port
        self.pop_ssl = pop_ssl
        self.pop_tls = pop_tls

        self.auto_add_from = auto_add_from
        self.auto_add_to = auto_add_to

        self.smtp_server = None  # type:SMTPServer
        self.pop_server = None  # type:POPServer

        # Check arguments.
        if not isinstance(self.log, logging.Logger):
            raise InvalidArguments('log excepted type logging.Logger got {}'.format(type(self.log)))

        if not isinstance(self.timeout, (int, float)):
            raise InvalidArguments('timeout excepted type int or float got {}'.format(type(self.timeout)))

        self.prepare()

    def prepare(self):
        """Init SMTPServer and POPServer."""
        if self.smtp_server is None:
            self.smtp_server = SMTPServer(username=self.username,
                                          password=self.password,
                                          host=self.smtp_host,
                                          port=self.smtp_port,
                                          ssl=self.smtp_ssl,
                                          tls=self.smtp_tls,
                                          timeout=self.timeout,
                                          debug=self.debug,
                                          log=self.log)
        if self.pop_server is None:
            self.pop_server = POPServer(username=self.username,
                                        password=self.password,
                                        host=self.pop_host,
                                        port=self.pop_port,
                                        ssl=self.pop_ssl,
                                        tls=self.pop_tls,
                                        timeout=self.timeout,
                                        debug=self.debug,
                                        log=self.log)

    def send_mail(self, recipients: List[str] or str, mail: dict or CaseInsensitiveDict, timeout=None,
                  auto_add_from=True, auto_add_to=True, cc=None) -> bool:
        """"Send email."""
        _mail = Mail(mail, debug=self.debug, log=self.log)

        if first_not_none(auto_add_from, self.auto_add_from) and _mail.mail.get('From') is None:
            _mail.set_mime_header('From', make_address_header([self.username]))

        recipients = make_list(recipients)
        if first_not_none(auto_add_to, self.auto_add_to) and _mail.mail.get('To') is None:
            _mail.set_mime_header('To', make_address_header(recipients))

        # Add Carbon Copy address.
        cc = make_list(cc) if cc is not None else None
        if cc is not None:
            for address in cc:
                recipients.append(address)
            _mail.set_mime_header('Cc', make_address_header(cc))

        # Remove tuple format in recipients.
        recipients = [i if not isinstance(i, tuple) else i[1] for i in recipients]

        with self.smtp_server as server:
            server.send(recipients, _mail,
                        first_not_none(timeout, self.timeout))

        return True

    def delete(self, which: int) -> bool:
        """Delete mail."""
        with self.pop_server as server:
            server.delete(which)
        return True

    def stat(self) -> tuple:
        """Get mailbox status."""
        with self.pop_server as server:
            return server.stat()

    def get_mail(self, which: int) -> CaseInsensitiveDict:
        """Get a mail from mailbox."""
        with self.pop_server as server:
            mail = server.get_mail(which)
            return parse_mail(mail, which, self.debug, self.log)

    def get_mails(self, subject=None, start_time=None, end_time=None, sender=None,
                  start_index: Optional[int] = None, end_index: Optional[int] = None) -> list:
        """Get a list of mails from mailbox."""
        headers = self.get_headers(start_index, end_index)
        mail_id = []

        if start_time is not None:
            if isinstance(start_time, (datetime.datetime, str)):
                start_time = convert_date_to_datetime(start_time)
            else:
                raise InvalidArguments(
                    'start_time excepted type str or datetime.datetime, got {} instead.'.format(type(start_time)))

        if end_time is not None:
            if isinstance(end_time, (datetime.datetime, str)):
                end_time = convert_date_to_datetime(end_time)
            else:
                raise InvalidArguments(
                    'end_time excepted type str or datetime.datetime, got {} instead.'.format(type(end_time)))

        for header in headers:
            if match_conditions(header, subject, start_time, end_time, sender):
                mail_id.append(header['id'])

        with self.pop_server as server:
            mail_id.sort()
            mail_as_bytes_list = server.get_mails(mail_id)
            return [parse_mail(mail_as_bytes, mail_id[index], self.debug, self.log)
                    for index, mail_as_bytes in enumerate(mail_as_bytes_list)]

    def get_latest(self) -> CaseInsensitiveDict:
        """Get latest mail in mailbox."""
        with self.pop_server as server:
            latest_num = server.stat()[0]
            mail = server.get_mail(latest_num)
            return parse_mail(mail, latest_num, self.debug, self.log)

    def get_info(self) -> List[List[bytes]]:
        warnings.warn("server.get_info is deprecated, if you want to access mail headers,"
                      "use server.get_headers instead",
                      DeprecationWarning,
                      stacklevel=2)
        with self.pop_server as server:
            return server.get_headers()

    def get_headers(self, start_index: Optional[int] = None, end_index: Optional[int] = None) \
            -> List[CaseInsensitiveDict]:
        """Get mails headers."""
        headers = []

        with self.pop_server as server:
            end = server.stat()[0]
            intersection = get_intersection((1, end), (start_index, end_index))  # type:List[int]
            mail_hdrs = server.get_headers(intersection)

        for index, mail_header in enumerate(mail_hdrs):
            _, _headers, *__ = parse_headers(mail_header)
            _headers.update(id=intersection[index])
            headers.append(_headers)

        return headers

    def log_debug(self, *args, **kwargs):
        self.log.debug(*args, **kwargs)

    def log_exception(self, *args, **kwargs):
        self.log_exception(*args, **kwargs)

    def smtp_able(self) -> bool:
        return self.smtp_server.check_available()

    def pop_able(self) -> bool:
        return self.pop_server.check_available()


class SMTPServer(BaseServer):
    """Base SMTPServer, which encapsulates python3 standard library to a SMTPServer."""

    def _make_server(self):
        """Init Server if possible."""
        if self.server is None:
            if self.ssl:
                self.server = smtplib.SMTP_SSL(self.host, self.port, __local__, timeout=self.timeout)
            else:
                self.server = smtplib.SMTP(self.host, self.port, __local__, timeout=self.timeout)

    def _remove_server(self):
        self.server = None

    def login(self):
        if self._login:
            self.log_exception('{} duplicate login!'.format(self.__repr__()))
            return

        if self.debug:
            self.log_access('login')

        self._make_server()

        if self.tls:
            self.stls()

        self.server.login(self.username, self.password)

        self._login = True

    def logout(self):
        if not self._login:
            self.log_exception('{} Logout before login!'.format(self.__repr__()))
            return

        if self.debug:
            self.log_access('logout')

        # Copied from smtplib.SMTP.__exit__
        # used for close connection.
        try:
            code, message = self.server.docmd("QUIT")
            if code != 221:
                raise smtplib.SMTPResponseException(code, message)
        except smtplib.SMTPServerDisconnected:
            pass
        finally:
            self.server.close()

        self._remove_server()

        self._login = False

    def stls(self):
        """Start TLS."""
        self.server.ehlo()
        self.server.starttls()
        self.server.ehlo()

    # Methods
    def send(self, recipients: Iterable[str], mail: Mail,
             timeout: int or float or None):

        if timeout is not None:
            self.server.timeout = timeout

        self.server.sendmail(self.username, recipients, mail.get_mime_as_string())


class POPServer(BaseServer):
    """Base POPServer, which encapsulates python3 standard library to a POPServer."""

    def _make_server(self):
        """Init Server."""
        if self.server is None:
            if self.ssl:
                self.server = poplib.POP3_SSL(self.host, self.port, timeout=self.timeout)
            else:
                self.server = poplib.POP3(self.host, self.port, timeout=self.timeout)

    def _remove_server(self):
        self.server = None

    def login(self):
        """Note: the mailbox on the server is locked until logout() is called."""
        if self._login:
            self.log_exception('{} duplicate login!'.format(self.__repr__()))
            return

        if self.debug:
            self.log_access('login')

        self._make_server()

        if self.tls:
            self.stls()

        self.server.user(self.username)
        self.server.pass_(self.password)

        self._login = True

    def logout(self):
        """Quit and remove pop3 server."""
        if not self._login:
            self.log_exception('{} Logout before login!'.format(self.__repr__()))
            return

        if self.debug:
            self.log_access('logout')

        self.server.quit()

        self._remove_server()

        self._login = False

    def stls(self):
        self.server.stls()

    # Methods

    def stat(self) -> tuple:
        """Get mailbox status. The result is a tuple of 2 integers: (message count, mailbox size)."""
        return self.server.stat()

    def get_header(self, which: int) -> list:
        """Use 'top' to get mail headers."""
        return self.server.top(which, 0)[1]

    def get_headers(self, which_list: Optional[list] = None) -> list:
        """Get all mails headers."""
        num = self.stat()[0]
        result = []

        if which_list is None:
            _range = range(1, num + 1)
        else:
            _range = which_list

        for count in _range:
            header_as_bytes = self.get_header(count)
            result.append(header_as_bytes)

        return result

    def get_mail(self, which: int) -> list:
        """Get a mail by its id."""
        return self.server.retr(which)[1]

    def get_mails(self, which_list: list) -> list:
        """Get a list of mails by its id."""
        return [self.server.retr(which)[1] for which in which_list]

    def delete(self, which: int):
        self.server.dele(which)



"""
zmail.settings
~~~~~~~~~~~~
This module contains some settings for zmail running or test.
"""

__publish__ = True

__local__ = 'zmail.local'

# Logging level: INFO == 20 , WARNING == 30 , ERROR == 40 , CRITICAL == 50
__level__ = 30 if __publish__ else 20



import json
import os

import pytest


@pytest.fixture(scope='module')
def here():
    return os.path.abspath(os.path.dirname(__file__))


@pytest.fixture(scope='module')
def accounts(here):
    accounts_path = os.path.join(here, 'config')

    with open(accounts_path, 'r') as f:
        raw = f.read()

    return json.loads(raw)



"""
zmail.info
~~~~~~~~~~
This module provide supported server information.


'Server_provider_address':{
    'protocol':('protocol_server_address', port, use_ssl,use_tls),
}
"""
from typing import Optional

SUPPORTED_SERVER = {
    '163.com': {
        'smtp_host': 'smtp.163.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.163.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.163.com',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    '126.com': {
        'smtp_host': 'smtp.126.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.126.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.126.com',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    'yeah.net': {
        'smtp_host': 'smtp.yeah.net',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.yeah.net',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.yeah.net',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    'qq.com': {
        'smtp_host': 'smtp.qq.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.qq.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'gmail.com': {
        'smtp_host': 'smtp.gmail.com',
        'smtp_port': 587,
        'smtp_ssl': False,
        'smtp_tls': True,
        'pop_host': 'pop.gmail.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'sina.com': {
        'smtp_host': 'smtp.sina.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.sina.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'outlook.com': {
        'smtp_host': 'smtp-mail.outlook.com',
        'smtp_port': 587,
        'smtp_ssl': False,
        'smtp_tls': True,
        'pop_host': 'pop.outlook.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'hotmail.com': {
        'smtp_host': 'smtp.office365.com',
        'smtp_port': 587,
        'smtp_ssl': False,
        'smtp_tls': True,
        'pop_host': 'outlook.office365.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
}

SUPPORTED_ENTERPRISE_SERVER_CONFIG = {
    'qq': {
        'smtp_host': 'smtp.exmail.qq.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.exmail.qq.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },
    'ali': {
        'smtp_host': 'smtp.mxhichina.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop3.mxhichina.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },
    '163': {
        'smtp_host': 'smtp.qiye.163.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.qiye.163.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },

}

DEFAULT_SERVER_CONFIG = {
    'smtp_host': 'smtp.',
    'smtp_port': 465,
    'smtp_ssl': True,
    'smtp_tls': False,
    'pop_host': 'pop.',
    'pop_port': 995,
    'pop_ssl': True,
    'pop_tls': False,
    'imap_host': 'imap.',
    'imap_port': 993,
    'imap_ssl': True,
    'imap_tls': False
}


def get_supported_server_info(mail_address: str, config: Optional[str] = None) -> dict:
    """Use user address to get server address and port."""
    provider = mail_address.split('@')[1]

    if config is not None:
        if config in SUPPORTED_ENTERPRISE_SERVER_CONFIG:
            return SUPPORTED_ENTERPRISE_SERVER_CONFIG[config]
        else:
            raise RuntimeError('Can not this config "{}".'.format(config))

    if provider in SUPPORTED_SERVER:
        return SUPPORTED_SERVER[provider]
    else:
        # Return default configs.
        config = DEFAULT_SERVER_CONFIG.copy()
        config['smtp_host'] += provider
        config['pop_host'] += provider
        config['imap_host'] += provider
        return config



import datetime
import os
import re
from base64 import b64encode
from typing import Optional

from .exceptions import InvalidArguments, ZmailInternalError
from .structures import CaseInsensitiveDict

DATETIME_PATTERN = re.compile(r'([0-9]+)?-?([0-9]{1,2})?-?([0-9]+)?\s*([0-9]{1,2})?:?([0-9]{1,2})?:?([0-9]{1,2})?\s*')
HDR_PREV = '=?utf-8?b?'
HDR_END = '?='


def convert_date_to_datetime(_date: str or datetime.datetime) -> datetime.datetime:
    """Convert date like '2018-1-1 12:00:00' to datetime object."""
    if isinstance(_date, datetime.datetime):
        # Shortcut
        return _date

    _match_info = DATETIME_PATTERN.fullmatch(_date)
    if _match_info is not None:
        year, month, day, hour, minute, second = [int(i) if i is not None else None for i in _match_info.groups()]
    else:
        raise InvalidArguments('Invalid date format ' + str(_date))

    if None in (year, month, day):
        now = datetime.datetime.now()
        year = year or now.year
        month = month or now.month
        day = day or now.day
    if None in (hour, minute, second):
        hour = hour or 0
        minute = minute or 0
        second = second or 0

    return datetime.datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second)


def match_conditions(mail_headers: CaseInsensitiveDict,
                     subject: Optional[str] = None,
                     start_time: Optional[datetime.datetime] = None,
                     end_time: Optional[datetime.datetime] = None,
                     sender: Optional[str] = None) -> bool:
    """Match all conditions."""
    mail_subject = mail_headers.get('subject')  # type:str or None
    mail_sender = mail_headers.get('from')  # type:str or None
    mail_date = mail_headers.get('date')  # type: datetime.datetime or None

    if subject is not None:
        if mail_subject is None or subject not in mail_subject:
            return False

    if sender is not None:
        if mail_sender is None or sender not in mail_sender:
            return False

    if start_time is not None:
        if mail_date is None or start_time > mail_date:
            return False

    if end_time is not None:
        if mail_date is None or end_time < mail_date:
            return False

    return True


def get_intersection(main_range: tuple, sub_range: tuple) -> list:
    main_start, main_end = main_range
    sub_start, sub_end = sub_range

    if main_start > main_end:
        return list()

    if sub_start is None or sub_start < main_start:
        sub_start = main_start
    if sub_end is None or sub_end > main_end:
        sub_end = main_end

    main_set = {i for i in range(main_start, main_end + 1)}
    sub_set = {i for i in range(sub_start, sub_end + 1)}

    return sorted(tuple((main_set & sub_set)))


def encode_mail_header(s: str) -> str:
    if not s:
        return ''
    return HDR_PREV + b64encode(s.encode('utf-8')).decode('ascii') + HDR_END


def make_list(obj) -> list:
    return obj if isinstance(obj, list) else [obj]


def make_address_header(address_list: list) -> str:
    """Used for make 'To' 'Cc' 'From' header."""
    res = []
    for address in address_list:
        if isinstance(address, tuple):
            assert len(address) == 2, 'Only two arguments!'
            name, rel_address = address
            res.append(encode_mail_header(name) + ' ' + '<{}>'.format(rel_address))
        elif isinstance(address, str):
            res.append('<{}>'.format(address))
        else:
            raise InvalidArguments('Email address can only be tuple or str.Get {} instead.'.format(type(address)))

    return ', '.join(res)


def first_not_none(*args):
    """Return first arguments which is not None."""
    for k in args:
        if k is not None:
            return k
    raise ZmailInternalError('All arguments is None!')


def get_abs_path(file: str) -> str:
    """if the file exists, return its abspath or raise a exception."""
    if os.path.exists(file):
        return file

    # Assert file exists in currently directory.
    work_path = os.path.abspath(os.getcwd())

    if os.path.exists(os.path.join(work_path, file)):
        return os.path.join(work_path, file)
    else:
        raise FileExistsError("The file %s doesn't exist." % file)



import datetime
import logging
import os
import time

import pytest

from zmail.exceptions import InvalidArguments
from zmail.info import get_supported_server_info
from zmail.server import MailServer, POPServer, SMTPServer

logger = logging.getLogger('zmail')


@pytest.fixture
def mail_server_config(accounts) -> dict:
    account = accounts[0]
    username = account[0]
    password = account[1]

    auto_generate_config = get_supported_server_info(username)

    # Ignore IMAP config.
    auto_generate_config = {k: v for k, v in auto_generate_config.items() if 'imap' not in k}

    auto_generate_config.update(username=username, password=password)

    return auto_generate_config


@pytest.fixture
def mail_server(mail_server_config: dict) -> MailServer:
    return MailServer(**mail_server_config)


def test_initiate_server(mail_server_config):
    config = mail_server_config
    server = MailServer(**config)
    for k, v in config.items():
        assert getattr(server, k) is v

    assert isinstance(server.smtp_server, SMTPServer)
    assert isinstance(server.pop_server, POPServer)
    assert server.log is logger

    # Test user-defined arguments.
    config_defined = config.copy()
    mock_logger = logging.Logger('mock')
    config_defined.update(timeout=30, debug=True, log=mock_logger, auto_add_from=False,
                          auto_add_to=False)
    server_defined = MailServer(**config_defined)
    assert server_defined.timeout == 30
    assert server_defined.debug is True
    assert server_defined.log is mock_logger
    assert server_defined.auto_add_from is False
    assert server_defined.auto_add_to is False

    with pytest.raises(InvalidArguments):
        config_with_error_logger = config.copy()
        config_with_error_logger.update(log=NotImplemented)
        MailServer(**config_with_error_logger)

    with pytest.raises(InvalidArguments):
        config_with_error_timeout = config.copy()
        config_with_error_timeout.update(timeout=NotImplemented)
        MailServer(**config_with_error_timeout)


def test_smtp_able_and_pop_able(mail_server: MailServer):
    assert mail_server.smtp_able()
    assert mail_server.pop_able()


def test_stat(mail_server: MailServer):
    stat = mail_server.stat()

    assert isinstance(stat, tuple)
    assert isinstance(stat[0], int)
    assert isinstance(stat[1], int)


def test_send_and_get(mail_server: MailServer, here):
    datetime_stamp = str(datetime.datetime.now())
    mail_as_dict = {
        'subject': '新测试邮件test! ' + datetime_stamp,
        'content_text': ['测试内容1_1\r\n', '测试内容1_2'],
        'content_html': '<html> \n<body>\n <h1>标题</h1>\n</body>',
        'from': 'ZMAIL测试',
        'to': 'ZMAIL测试',
        'attachments': [os.path.join(here, '图标.ico'), os.path.join(here, 'favicon.ico')]
    }

    parsed_attachments = [('图标.ico',
                           b'\x00\x00\x01\x00\x01\x00\x10\x10\x00\x00\x01\x00 \x00h\x04\x00\x00\x16\x00\x00\x00(\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'),
                          ('favicon.ico',
                           b'\x00\x00\x01\x00\x01\x00\x10\x10\x00\x00\x01\x00 \x00h\x04\x00\x00\x16\x00\x00\x00(\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\xff\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]

    # Test send and get_mail().
    now_latest_num = mail_server.stat()[0]
    mail_server.send_mail(mail_server.username, mail_as_dict)
    t = 0
    while True:
        if now_latest_num < mail_server.stat()[0]:
            mail_receive = mail_server.get_mail(now_latest_num + 1)
            break
        else:
            t += 1
            if t == 120:
                raise TimeoutError('Test timeout({}s).'.format(t * 0.5))
            time.sleep(0.5)
    for k in mail_as_dict:
        if k != 'attachments':
            assert mail_receive[k] == mail_receive[k]
        else:
            assert mail_receive[k] == parsed_attachments

    # Test get_latest()
    assert mail_server.get_latest() == mail_server.get_mail(now_latest_num + 1)

    # Test get_mails()
    mail_receive = mail_server.get_mails(subject=mail_as_dict['subject'],
                                         start_time='1970-1-1', end_time='2200-1-1',
                                         sender='ZMAIL测试')[0]

    mail_receive_with_index = mail_server.get_mails(subject=mail_as_dict['subject'],
                                                    start_time='1970-1-1', end_time='2200-1-1',
                                                    sender='ZMAIL测试', start_index=now_latest_num)[0]

    assert mail_receive == mail_receive_with_index

    for k in mail_as_dict:
        if k != 'attachments':
            assert mail_receive[k] == mail_receive[k]
        else:
            assert mail_receive[k] == parsed_attachments

    with pytest.raises(InvalidArguments):
        mail_server.get_mails(start_time=b'test')
    with pytest.raises(InvalidArguments):
        mail_server.get_mails(end_time=b'test')
    with pytest.raises(InvalidArguments):
        mail_server.get_mails(start_time=b'test', end_time=b'test')

    # Test delete()
    assert mail_server.delete(mail_receive['id'])
    assert mail_server.stat()[0] < mail_receive['id'] or mail_server.get_mail(mail_receive['id']) != mail_receive


def test_send_cc(mail_server: MailServer):
    mail_as_dict = {
        'subject': 'test_cc',
        'content_text': ['测试内容1_1\r\n', '测试内容1_2'],
    }

    now_latest_num = mail_server.stat()[0]
    mail_server.send_mail(mail_server.username, mail_as_dict, cc=mail_server.username)
    t = 0
    while True:
        if now_latest_num < mail_server.stat()[0]:
            mail_receive = mail_server.get_mail(now_latest_num + 1)
            break
        else:
            t += 1
            if t == 120:
                raise TimeoutError('Test timeout({}s).'.format(t * 0.5))
            time.sleep(0.5)

    for k in mail_as_dict:
        assert mail_receive[k] == mail_receive[k]

    assert mail_receive['headers']['Cc'] == '<{}>'.format(mail_server.username)
    mail_server.delete(mail_receive['id'])


def test_resend(mail_server: MailServer, here):
    mail_as_dict = {'subject': str(datetime.datetime.now()),
                    'attachments': os.path.join(here, 'favicon.ico'),
                    'content_text': ['111', '222'],
                    'content_html': ['222', '333']}
    # Send first mail.
    now_latest_num = mail_server.stat()[0]
    mail_server.send_mail(mail_server.username, mail_as_dict)
    t = 0
    while True:
        if now_latest_num < mail_server.stat()[0]:
            mail_receive = mail_server.get_mail(now_latest_num + 1)
            break
        else:
            t += 1
            if t == 120:
                raise TimeoutError('Test timeout({}s).'.format(t * 0.5))
            time.sleep(0.5)

    # Send received mail.
    now_latest_num = mail_server.stat()[0]
    mail_server.send_mail(mail_server.username, mail_receive)
    t = 0
    while True:
        if now_latest_num < mail_server.stat()[0]:
            mail_receive_next = mail_server.get_mail(now_latest_num + 1)
            break
        else:
            t += 1
            if t == 120:
                raise TimeoutError('Test timeout({}s).'.format(t * 0.5))
            time.sleep(0.5)
    for k in ('subject', 'from', 'to', 'content_text', 'content_html', 'attachments'):
        assert mail_receive[k] == mail_receive_next[k]

    # Delete these mails.
    mail_server.delete(mail_receive_next['id'])
    mail_server.delete(mail_receive['id'])


def test_all_account_smtp_and_pop_able(accounts):
    for account in accounts:
        username = account[0]
        password = account[1]

        auto_generate_config = get_supported_server_info(username)

        # Ignore IMAP config.
        auto_generate_config = {k: v for k, v in auto_generate_config.items() if 'imap' not in k}

        auto_generate_config.update(username=username, password=password)

        server = MailServer(**auto_generate_config)

        assert server.smtp_able()
        assert server.pop_able()



import logging
from abc import ABC, abstractmethod
from typing import Optional

logger = logging.getLogger('zmail')


class BaseServer(ABC):
    """Base protocol server."""

    def __init__(self, username: str, password: str,
                 host: str, port: int,
                 ssl: bool, tls: bool,
                 timeout: int or float,
                 debug: bool, log: Optional[logging.Logger] = None):
        self.server = None
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.timeout = timeout

        self.ssl = ssl
        self.tls = tls
        self.debug = debug
        self.log = log or logger

        self._login = False

        if tls and ssl:
            raise TypeError('Can not use ssl and tls together.')

    @abstractmethod
    def _make_server(self):
        pass

    def _remove_server(self):
        self.server = None

    @abstractmethod
    def login(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    @abstractmethod
    def stls(self):
        pass

    def check_available(self) -> bool:
        try:
            self.login()
            self.logout()
            return True
        except Exception as e:
            self.log_exception('{} access error :{}'.format(self.__class__.__name__, e))
            return False

    def is_login(self) -> bool:
        return self._login

    def log_debug(self, msg):
        self.log.debug(msg)

    def log_exception(self, msg):
        self.log.fatal(msg)

    def log_access(self, msg):
        self.log_debug('<{} {}:{} ssl:{} tls:{} is_login:{}> {}.'
                       .format(self.__class__.__name__,
                               self.host, self.port,
                               self.ssl, self.tls,
                               self.is_login(), msg))

    def __enter__(self):
        self.login()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logout()

    def __repr__(self):
        return '<{} username:{} password:{} ' \
               '{}:{} ssl:{} tls:{}>' \
            .format(self.__class__.__name__, self.username, self.password,
                    self.host, self.port, self.ssl, self.tls)



import os
from contextlib import suppress

import pytest

from zmail.api import server
from zmail.structures import CaseInsensitiveDict
from zmail.utils import read, save, save_attachment


def test_save_attachment(here):
    mock_mail = CaseInsensitiveDict({
        'attachments': [('_test_0.txt', b'55555'), ('_test_1.txt', b'66666')]
    })
    file_name_0 = mock_mail['attachments'][0][0]
    file_raw_0 = mock_mail['attachments'][0][1]
    file_name_1 = mock_mail['attachments'][1][0]
    file_raw_1 = mock_mail['attachments'][1][1]

    path_0 = os.path.join(here, file_name_0)  # Mock user defined path.
    path_1 = os.path.join(here, file_name_1)
    try:
        save_attachment(mock_mail)

        with open(file_name_0, 'rb') as f:
            assert f.read() == file_raw_0

        with open(file_name_1, 'rb') as f:
            assert f.read() == file_raw_1

        save_attachment(mock_mail, overwrite=True)  # No FileExistsError
        with pytest.raises(FileExistsError):
            save_attachment(mock_mail)

        # Test defined path.
        save_attachment(mock_mail, here)

        with open(path_0, 'rb') as f:
            assert f.read() == file_raw_0

        with open(path_1, 'rb') as f:
            assert f.read() == file_raw_1

        save_attachment(mock_mail, here, overwrite=True)  # No FileExistsError
        with pytest.raises(FileExistsError):
            save_attachment(mock_mail, here)
    except (BaseException, Exception):
        raise
    finally:
        with suppress(FileNotFoundError):
            os.remove(file_name_0)
            os.remove(file_name_1)
            os.remove(path_0)
            os.remove(path_1)


def test_save_and_read(accounts):
    try:
        account = accounts[0]
        username, password = account
        if server(username, password).stat()[0]:
            mail = server(username, password).get_latest()
            save(mail, name='_test.eml')
            saved_mail = read('_test.eml')
            saved_mail['id'] = mail.get('id')
            assert mail == saved_mail
    except (BaseException, Exception):
        raise
    finally:
        with suppress(FileNotFoundError):
            os.remove('_test.eml')



import logging
import mimetypes
import os
import warnings
from email.encoders import encode_base64
from email.header import Header
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List, Optional

from .exceptions import InvalidArguments
from .helpers import get_abs_path, make_list
from .parser import parse
from .structures import CaseInsensitiveDict

logger = logging.getLogger('zmail')


class Mail:
    def __init__(self, mail: dict or CaseInsensitiveDict, boundary: Optional[str] = None,
                 debug: bool = False, log: logging.Logger = None):
        if isinstance(mail, dict):
            self.mail = CaseInsensitiveDict(mail)
        elif isinstance(mail, CaseInsensitiveDict):
            self.mail = mail
        else:
            raise InvalidArguments('mail field excepted type dict or CaseInsensitiveDict got {}'.format(type(mail)))

        self.boundary = boundary
        self.debug = debug
        self.log = log or logger
        self.mime = None  # type:MIMEMultipart

    def make_mine(self) -> None:
        mime = MIMEMultipart(boundary=self.boundary)

        # Set basic email elements.
        for k, v in self.mail.items():
            _k = k.lower()
            if _k in ('from', 'to', 'subject') and v is not None:
                mime[_k.capitalize()] = v
            elif _k in ('attachments', 'content_text', 'content_html'):
                pass
            elif _k == 'headers':
                pass
            else:
                if not all([(i in self.mail) for i in
                            ('from', 'to', 'subject', 'raw_headers', 'charsets', 'headers',
                             'date', 'id', 'raw', 'attachments', 'content_text', 'content_html')]):
                    # Remove resend warnings.
                    warnings.warn("Header '{}' is invalid and unused,if you want to add extra headers "
                                  "use 'headers' instead.".format(str(_k)), category=DeprecationWarning, stacklevel=4)

        # Set extra headers.
        if self.mail.get('headers') and isinstance(self.mail['headers'], dict):
            for k, v in self.mail['headers'].items():
                mime[k] = v

        # Set HTML content.
        if self.mail.get('content_html') is not None:
            _htmls = make_list(self.mail['content_html'])
            for _html in _htmls:
                mime.attach(MIMEText('{}'.format(_html), 'html', 'utf-8'))

        # Set TEXT content.
        if self.mail.get('content_text') is not None:
            _messages = make_list(self.mail['content_text'])
            for _message in _messages:
                mime.attach(MIMEText('{}'.format(_message), 'plain', 'utf-8'))

        # Set attachments.
        if self.mail.get('attachments'):
            attachments = make_list(self.mail['attachments'])
            for attachment in attachments:
                if isinstance(attachment, str):
                    attachment_abs_path = get_abs_path(attachment)
                    part = make_attachment_part(attachment_abs_path)
                    mime.attach(part)
                elif isinstance(attachment, tuple):
                    name, raw = attachment
                    part = MIMEBase('application', 'octet-stream')
                    part.set_payload(raw)
                    part['Content-Disposition'] = 'attachment;filename="{}"'.format(name)
                    encode_base64(part)
                    mime.attach(part)
                else:
                    raise InvalidArguments('Attachments excepted str or tuple got {} instead.'.format(type(attachment)))

        self.mime = mime

    def set_mime_header(self, k, v) -> None:
        if self.mime is not None:
            self.mime[k] = v
        else:
            self.make_mine()
            self.mime[k] = v

    def decode(self) -> CaseInsensitiveDict:
        if self.mime is None:
            self.make_mine()
        return parse(self.mime.as_string().encode('utf-8').split(b'\n'))

    def get_mime_raw(self) -> MIMEMultipart:
        if self.mime is not None:
            return self.mime
        else:
            self.make_mine()
            return self.mime

    def get_mime_as_string(self) -> str:
        return self.get_mime_raw().as_string()

    def get_mime_as_bytes_list(self) -> List[bytes]:
        return self.get_mime_as_string().encode('utf-8').split(b'\n')


def make_attachment_part(file_path) -> MIMEBase:
    """According to file-type return a prepared attachment part."""
    name = os.path.split(file_path)[1]
    file_type = mimetypes.guess_type(name)[0]

    encoded_name = Header(name).encode()

    if file_type is None:
        logger.warning('Could not guess %s type, use application type instead.', file_path)
        file_type = 'application/octet-stream'

    main_type, sub_type = file_type.split('/')

    if main_type == 'text':
        with open(file_path, 'r') as f:
            part = MIMEText(f.read())
            part['Content-Disposition'] = 'attachment;filename="{}"'.format(encoded_name)

    elif main_type in ('image', 'audio'):
        with open(file_path, 'rb') as f:
            part = MIMEImage(f.read(), _subtype=sub_type) if main_type == 'image' else \
                MIMEAudio(f.read(), _subtype=sub_type)
            part['Content-Disposition'] = 'attachment;filename="{}"'.format(encoded_name)

    else:
        with open(file_path, 'rb') as f:
            part = MIMEBase(main_type, sub_type)
            part.set_payload(f.read())
            part['Content-Disposition'] = 'attachment;filename="{}"'.format(encoded_name)
            encode_base64(part)
    return part



"""
zmail.__init__
~~~~~~~~~~~~
Zmail allows you to send and get email as possible as it can be.
"""
import logging
from .api import *  # noqa
from .settings import __level__  # noqa

# Define logger.
logger = logging.getLogger('zmail')

sh = logging.StreamHandler()
fmt = logging.Formatter('[%(levelname)s] %(message)s')
sh.setFormatter(fmt)
logger.addHandler(sh)

logger.setLevel(logging.DEBUG)

# A standard zmail dict.
MAIL = {
    'subject': 'Zmail',  # Anything you want.
    'content': 'This message from zmail!',  # Anything you want.
    'attachments': '',  # Absolute path will be better.
}



"""
zmail.structures
~~~~~~~~~~~~~~~~
Data structures that power zmail.
"""
import collections


class CaseInsensitiveDict(collections.MutableMapping):
    """A case-insensitive ``dict``-like object.
    Inspired by requests.structures

    _store -> 'key.lower()' : ('key','value')
    """

    def __init__(self, data=None, **kwargs):
        self._store = collections.OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, _ in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lowerkey, value[1])
            for (lowerkey, value)
            in self._store.items()
        )

    def __eq__(self, other):
        if isinstance(other, collections.Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    def copy(self):
        """Shallow copy."""
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))



"""
zmail.utils
~~~~~~~~~~~~
This module contains some useful function power zmail.
"""

import os
from typing import Optional

from .helpers import get_abs_path, make_list
from .parser import parse_mail
from .structures import CaseInsensitiveDict


def save_attachment(mail: CaseInsensitiveDict, target_path: Optional[str] = None, overwrite=False):
    """Parsing attachment and save it."""
    if mail.get('attachments'):
        if target_path is not None:
            assert os.path.isdir(target_path) and os.path.exists(target_path)
        else:
            target_path = os.getcwd()

        for name, raw in mail['attachments']:
            file_path = os.path.join(target_path, name)
            if not overwrite and os.path.exists(file_path):
                raise FileExistsError("{} already exists, set overwrite to True to avoid this error.")
            with open(file_path, 'wb') as f:
                f.write(raw)


def show(mails: list or CaseInsensitiveDict) -> None:
    """Show mail or mails."""
    mails = make_list(mails)
    for mail in mails:
        print('-------------------------')
        for k in ('subject', 'id', 'from', 'to', 'date', 'content_text', 'content_html', 'attachments'):
            if k != 'attachments':
                print(k.capitalize() + ' ', mail.get(k))
            else:
                _ = ''
                for idx, v in enumerate(mail['attachments']):
                    _ += str(idx + 1) + '.' + 'Name:' + v[0] + ' ' + 'Size:' + str(len(v[1])) + ' '

                print(k.capitalize() + ' ', _)


def read_html(html_path: str):
    """Get html content by its path."""
    path = get_abs_path(html_path)

    with open(path, 'r') as f:
        content = f.read()

    return content


def read(file_path: str, SEP=b'\r\n') -> CaseInsensitiveDict:
    """Read a mail."""
    abs_path = get_abs_path(file_path)

    with open(abs_path, 'rb') as f:
        raw_lines = f.read().split(SEP)

    return parse_mail(raw_lines, 0)


def save(mail, name=None, target_path=None, overwrite=False) -> bool:
    """Save a mail."""
    if name is None:
        name = str(mail['subject'] + '.eml') if mail.get('subject') else 'Untitled'

    if target_path is None:
        target_path = os.getcwd()

    file_path = os.path.join(target_path, name)

    if not overwrite and os.path.exists(file_path):
        raise FileExistsError("{} already exists, set overwrite to True to avoid this error.")

    with open(file_path, 'wb') as f:
        f.write(b'\r\n'.join(mail['raw']))

    return True



"""
zmail.exceptions
~~~~~~~~~~~~~~~~
Include all exceptions used within zmail.
"""


class ZmailException(RuntimeError):
    """Base exception."""


class InvalidProtocol(ZmailException, ValueError):
    """Invalid protocol settings used."""


class ParseError(ZmailException):
    """Parse mail error."""


class InvalidArguments(ZmailException):
    """Invalid arguments."""


class ZmailInternalError(ZmailException):
    """Internal error."""



from zmail.parser import recursive_decode, remove_line_feed_and_whitespace


def test_recursive_decode():
    test_as_str = 'zmail测试信息'
    assert recursive_decode(test_as_str.encode('utf-8'), ('utf-8',)) == test_as_str
    assert recursive_decode(test_as_str.encode('gbk'), ('gbk', 'utf-8')) == test_as_str
    assert recursive_decode(test_as_str.encode('gbk'), ('utf-8', 'gbk')) == test_as_str
    assert recursive_decode(test_as_str.encode('gbk'), ('utf-8', 'gb2312', 'gbk')) == test_as_str

    assert recursive_decode(test_as_str.encode('utf-8'), tuple()) is None
    assert recursive_decode(test_as_str.encode('gbk'), ('utf-8',)) is None

    assert recursive_decode(test_as_str.encode('utf-8'), ('gbk',)) == 'zmail娴嬭瘯淇℃伅'


def test_remove_line_feed_and_whitespace():
    assert remove_line_feed_and_whitespace('gb2312') == 'gb2312'
    assert remove_line_feed_and_whitespace('nr123rn') == 'nr123rn'

    assert remove_line_feed_and_whitespace(' gb2312 ') == 'gb2312'
    assert remove_line_feed_and_whitespace('r gb2312 n') == 'r gb2312 n'

    assert remove_line_feed_and_whitespace('\r\ngb2312\r\n') == 'gb2312'
    assert remove_line_feed_and_whitespace('\r\n \r\ngb2312\r\n \r\n') == 'gb2312'

    assert remove_line_feed_and_whitespace(r' \r\ngb2312\r\n ') == 'gb2312'
    assert remove_line_feed_and_whitespace(' \r\n\\r\\ngb2312\\r\\n\r\n ') == 'gb2312'

    assert remove_line_feed_and_whitespace(r' \r\n\r\ngb2312\r\n\r\n ') == 'gb2312'
    assert remove_line_feed_and_whitespace('\r\n \r\n\r\ngb2312\r\n\r\n \r\n') == 'gb2312'



import pytest

from zmail.info import (DEFAULT_SERVER_CONFIG,
                        SUPPORTED_ENTERPRISE_SERVER_CONFIG, SUPPORTED_SERVER,
                        get_supported_server_info)


def test_supported_server():
    assert isinstance(SUPPORTED_SERVER, dict)
    assert isinstance(SUPPORTED_ENTERPRISE_SERVER_CONFIG, dict)

    combine = SUPPORTED_SERVER.copy()
    combine.update(SUPPORTED_ENTERPRISE_SERVER_CONFIG)

    for host in combine:
        host_info = combine[host]
        smtp_host = host_info['smtp_host']
        smtp_port = host_info['smtp_port']
        smtp_ssl = host_info['smtp_ssl']
        smtp_tls = host_info['smtp_tls']
        pop_host = host_info['pop_host']
        pop_port = host_info['pop_port']
        pop_ssl = host_info['pop_ssl']
        pop_tls = host_info['pop_tls']

        assert isinstance(smtp_host, str), smtp_host
        assert isinstance(pop_host, str), pop_host

        assert isinstance(smtp_port, int)
        assert isinstance(pop_port, int)

        assert not (smtp_ssl and smtp_tls)
        assert not (pop_ssl and pop_tls)


def test_get_supported_server_info():
    for k in SUPPORTED_SERVER:
        assert get_supported_server_info('zmailtest@' + k), SUPPORTED_SERVER[k]


def test_get_supported_server_info_as_enterprise_server():
    for k in SUPPORTED_ENTERPRISE_SERVER_CONFIG:
        assert get_supported_server_info('zmailtest@gmail.com', config=k), SUPPORTED_ENTERPRISE_SERVER_CONFIG[k]

    with pytest.raises(RuntimeError):
        get_supported_server_info('zmailtest@gmail.com', config='zmailtest')


def test_get_supported_server_info_as_default_config():
    assert get_supported_server_info('zmailtest@zmailtest.com'), DEFAULT_SERVER_CONFIG



import logging
import poplib
from unittest import mock

import pytest

from zmail.info import get_supported_server_info
from zmail.server import POPServer


@pytest.fixture
def pop_server(accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    config = get_supported_server_info(username)
    return POPServer(username, password,
                     host=config['pop_host'],
                     port=config['pop_port'],
                     ssl=config['pop_ssl'],
                     tls=config['pop_tls'],
                     timeout=60,
                     debug=False)


@pytest.fixture
def pop_server_config(accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    config = get_supported_server_info(username)
    return {
        'username': username,
        'password': password,
        'host': config['pop_host'],
        'port': config['pop_port'],
        'ssl': config['pop_ssl'],
        'tls': config['pop_tls'],
        'timeout': 60,
        'debug': False
    }


def test_pop_server_init(pop_server_config, accounts):
    account = accounts[0]
    username = account[0]
    password = account[1]
    new_logger = mock.Mock()
    srv = POPServer(**pop_server_config, log=new_logger)
    assert srv.username == username
    assert srv.password == password
    assert srv.host == pop_server_config['host']
    assert srv.port == pop_server_config['port']
    assert srv.ssl == pop_server_config['ssl']
    assert srv.tls == pop_server_config['tls']
    assert srv.timeout == 60
    assert srv.debug is False
    assert srv.log is new_logger


def test_pop_server_default_logger(pop_server_config):
    srv = POPServer(**pop_server_config)
    assert srv.log == logging.getLogger('zmail')


def test_pop_ssl_tls_error(pop_server_config):
    configs = pop_server_config.copy()
    configs.update(tls=True)
    with pytest.raises(TypeError):
        POPServer(**configs)


def test_pop_make_server(pop_server_config):
    srv = POPServer(**pop_server_config)
    srv._make_server()
    assert isinstance(srv.server, (poplib.POP3_SSL if srv.ssl else poplib.POP3))


def test_pop_login(pop_server_config):
    srv = POPServer(**pop_server_config)
    srv.login()
    assert srv.is_login() is True and srv._login is True
    srv.logout()


def test_pop_duplicate_login(pop_server_config):
    srv = POPServer(**pop_server_config)
    srv.login()
    srv.login()
    assert srv.is_login() is True and srv._login is True
    srv.logout()


def test_pop_logout(pop_server_config):
    srv = POPServer(**pop_server_config)
    srv.login()
    srv.logout()
    assert srv.is_login() is False and srv._login is False


def test_pop_context(pop_server: POPServer):
    assert pop_server.is_login() is False
    with pop_server as server:
        assert isinstance(server, POPServer)
        assert server.server is not None
        assert server.is_login() is True and server._login is True
    assert server.server is None
    assert server.is_login() is False and server._login is False


def test_pop_check_available(pop_server_config):
    assert POPServer(**pop_server_config).check_available()
    incorrect_config = pop_server_config.copy()
    incorrect_config.update(host='')
    assert POPServer(**incorrect_config).check_available() is False


# Methods

def test_stat(pop_server: POPServer):
    with pop_server as server:
        stat = server.stat()
        assert isinstance(stat, tuple)
        assert len(stat) == 2
        assert isinstance(stat[0], int) and isinstance(stat[1], int)


def test_get_header(pop_server: POPServer):
    with pop_server as server:
        if server.stat()[0]:
            header = server.get_header(1)
            assert isinstance(header, list)
            for i in header:
                assert isinstance(i, bytes)


def test_get_headers(pop_server: POPServer):
    with pop_server as server:
        total_num = server.stat()[0]
        if total_num >= 2:
            which_list = list(range(1, 5 if total_num >= 5 else total_num))
            headers = server.get_headers(which_list)
            assert isinstance(headers, list)
            for header in headers:
                assert isinstance(header, list)
                for i in header:
                    assert isinstance(i, bytes)


def test_get_mail(pop_server: POPServer):
    with pop_server as server:
        total_num = server.stat()[0]
        if total_num:
            mail = server.get_mail(1)
            assert isinstance(mail, list)
            for i in mail:
                assert isinstance(i, bytes)


def test_get_mails(pop_server: POPServer):
    with pop_server as server:
        total_num = server.stat()[0]
        if total_num >= 2:
            which_list = list(range(1, 5 if total_num >= 5 else total_num))
            mails = server.get_mails(which_list)
            assert isinstance(mails, list)
            for mail in mails:
                assert isinstance(mail, list)
                for i in mail:
                    assert isinstance(i, bytes)



import os
import sys

from setuptools import find_packages, setup

PROJECT_NAME = 'zmail'
MODULE_NAME = 'zmail'

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('pip3 install twine')
    os.system('pip3 install wheel')
    os.system('python3 setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    os.system('rm -rf build dist .egg zmail.egg-info')
    sys.exit()

# 'setup.py test' shortcut.
if sys.argv[-1] == 'test':
    os.system('pip3 install pytest')
    os.system('pytest tests/')
    sys.exit()

# 'setup.py isort' shortcut.
if sys.argv[-1] == 'isort':
    os.system('isort -rc *.py')
    os.system('isort -rc zmail/*.py')
    os.system('isort -rc tests/*.py')
    sys.exit()

# 'setup.py flake8' shortcut.
if sys.argv[-1] in ('flake8', 'flake'):
    os.system('flake8 --ignore=E501 zmail/*.py tests/*.py')
    sys.exit()

setup(
    name='zmail',
    version='0.2.1',

    author='ZYunH',
    author_email='zyunhjob@163.com',

    description='Zmail allows you to send and get emails as possible as it can be in python',
    long_description='Zmail allows you to send and get emails as possible as it can be in python.'
                     'There is no need to check server address or make your own MIME string.With zmail, '
                     'you only need to care about your mail content.',
    keywords='email python3 package',

    url='https://github.com/ZYunH/zmail',
    license="MIT Licence",

    platforms='all',
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ),

    packages=find_packages(),
    include_package_data=True,

)
